/*  ======================================================================  */
/*  ==========                                                  ==========  */
/*  ==========                P O L Y . C                       ==========  */
/*  ==========                                                  ==========  */
/*  ==========       Version  3.0  by  H.S.    Sept. 2000       ==========  */
/*  ==========                                                  ==========  */
/*  ======================================================================  */

/*
 *	Coord   get coordinates by reading them or converting weight input
 *	Rat	rational functions
 *	Vertex	computes Vertices and Faces 
 *      Polynf	normal form and symmetries
 *	-> class.x
 */

#include "Global.h"

#define FIRST_OC (0)
#define SHOW_NEW_CEq (1)
#define STATIC	static

#ifndef	CEQ_Nmax	
#define CEQ_Nmax        EQUA_Nmax
#endif

FILE *inFILE, *outFILE;

typedef struct {int ne; Equation e[CEQ_Nmax];}              CEqList;

int  GLZ_Start_Simplex(PolyPointList *_P, VertexNumList *_V, CEqList *_C);
int  IsGoodCEq(Equation *_E, PolyPointList *_P, VertexNumList *_V);

int  Lead_Vert(INCI x){
  int i=0; 
  while(!(INCI_M2(x))) {i++; x=INCI_D2(x);} 
  return i;
}

void GBU_Make_New_CEqs(PolyPointList *_P, VertexNumList *_V, CEqList *_C, 
			INCI *CEq_I){
  int i,j, Good_C_ne=_C->ne;
  STATIC CEqList Bad_C, Ugly_C;
  STATIC INCI Bad_C_I[CEQ_Nmax], Ugly_C_I[CEQ_Nmax];

#if     (SHOW_NEW_CEq)
  static int init; static clock_t CLOCK1; static time_t DATE1;
  if(!init){init=1; CLOCK1=clock(); DATE1=time(NULL);}
  printf("V=%d Ceq=%d",_V->nv,_C->ne);fflush(stdout);
#endif

  Bad_C.ne=Ugly_C.ne=_C->ne=0;

  for (i=0;i<Good_C_ne;i++){
    Long dist = Eval_Eq_on_V(&_C->e[i],_P->x[_V->v[_V->nv-1]],_P->n);
    CEq_I[i]=INCI_PN(CEq_I[i],dist);
    if (dist>0) {
      _C->e[_C->ne]=_C->e[i]; 
      CEq_I[_C->ne++]=CEq_I[i];}
    else if (dist<0) {
      Bad_C.e[Bad_C.ne]=_C->e[i]; 
      Bad_C_I[Bad_C.ne++]=CEq_I[i];}
    else {Ugly_C.e[Ugly_C.ne]=_C->e[i]; Ugly_C_I[Ugly_C.ne++]=CEq_I[i];}}

#if     (SHOW_NEW_CEq)
  /* for (j=0;j<_V->nv;j++) printf(" %d",First_Oc[j]); */
  printf("=%dg+%db+%du",_C->ne,Bad_C.ne,Ugly_C.ne);fflush(stdout);
#endif

  Good_C_ne=_C->ne;

  for (i=0;i<Ugly_C.ne;i++){
    _C->e[_C->ne]=Ugly_C.e[i]; 
    CEq_I[_C->ne++]=Ugly_C_I[i];}

  for (j=0;j<Good_C_ne;j++)
    for (i=Bad_C.ne-1;i>=0;i--){
      INCI New_Face=INCI_AND(Bad_C_I[i],CEq_I[j]);
      int k, LV;
      if (INCI_abs(New_Face)<_P->n-1) continue;
      LV=_V->nv-1-Lead_Vert(New_Face);
      for (k=0;k<Bad_C.ne;k++) 
	if (INCI_LE(New_Face,Bad_C_I[k])) if (k!=i) break;
      if (k!=Bad_C.ne) continue;
      for (k=0;k<_C->ne;k++) if (INCI_LE(New_Face,CEq_I[k]))
	if (k!=j) break;
      if (k!=_C->ne) continue;
      assert(_C->ne<CEQ_Nmax);
      CEq_I[_C->ne]=INCI_PN(INCI_D2(New_Face),0);
      _C->e[_C->ne]=EEV_To_Equation(&(Bad_C.e[i]),&(_C->e[j]),
				    _P->x[_V->v[_V->nv-1]],_P->n);
      assert(IsGoodCEq(&(_C->e[_C->ne++]),_P,_V));}

#if     (SHOW_NEW_CEq)
  {time_t DATE2=time(NULL); char sm[2]={'s',0}; 
  int Rs= (int)difftime(DATE2,DATE1);if(Rs>999){Rs/=60; *sm='m';}
  printf(" done: C.ne=%d  %d%s\n",_C->ne,Rs,sm);fflush(0);}
#endif
}

void FO_GBU_Make_New_CEqs(PolyPointList *_P, VertexNumList *_V, CEqList *_C, 
			INCI *CEq_I, int *First_Oc){
  int i,j, Good_C_ne=_C->ne, New_First_Oc[VERT_Nmax], Bad_First_Oc[VERT_Nmax];
  STATIC CEqList Bad_C, Ugly_C;
  STATIC INCI Bad_C_I[CEQ_Nmax], Ugly_C_I[CEQ_Nmax];

#if     (SHOW_NEW_CEq)
  static int init; static clock_t CLOCK1; static time_t DATE1;
  if(!init){init=1; CLOCK1=clock(); DATE1=time(NULL);}
  printf("V=%d Ceq=%d",_V->nv,_C->ne);fflush(stdout);
#endif

  First_Oc[_V->nv-1]=_C->ne;
  Bad_C.ne=Ugly_C.ne=_C->ne=0;

  for (j=0;j<_V->nv-1;j++) {
    New_First_Oc[j+1]=_C->ne;
    Bad_First_Oc[j+1]=Bad_C.ne;
    for (i=First_Oc[j];i<First_Oc[j+1];i++){
      Long dist = Eval_Eq_on_V(&_C->e[i],_P->x[_V->v[_V->nv-1]],_P->n);
      CEq_I[i]=INCI_PN(CEq_I[i],dist);
      if (dist>0) {
	New_First_Oc[j+1]++;
	_C->e[_C->ne]=_C->e[i]; 
	CEq_I[_C->ne++]=CEq_I[i];}
      else if (dist<0) {
	Bad_First_Oc[j+1]++;
	Bad_C.e[Bad_C.ne]=_C->e[i]; 
	Bad_C_I[Bad_C.ne++]=CEq_I[i];}
      else {Ugly_C.e[Ugly_C.ne]=_C->e[i]; Ugly_C_I[Ugly_C.ne++]=CEq_I[i];}}}

#if     (SHOW_NEW_CEq)
  /* for (j=0;j<_V->nv;j++) printf(" %d",First_Oc[j]); */
  printf("=%dg+%db+%du",_C->ne,Bad_C.ne,Ugly_C.ne);fflush(stdout);
#endif

  Good_C_ne=_C->ne;
  for (j=1;j<_V->nv;j++) First_Oc[j]=New_First_Oc[j];
  for (i=0;i<Ugly_C.ne;i++){
    _C->e[_C->ne]=Ugly_C.e[i]; 
    CEq_I[_C->ne++]=Ugly_C_I[i];}

  for (j=0;j<Good_C_ne;j++)
    for (i=Bad_C.ne-1;i>=0;i--){
      INCI New_Face=INCI_AND(Bad_C_I[i],CEq_I[j]);
      int k, LV;
      if (INCI_abs(New_Face)<_P->n-1) continue;
      LV=_V->nv-1-Lead_Vert(New_Face);
      for (k=Bad_First_Oc[LV];k<Bad_C.ne;k++) 
	if (INCI_LE(New_Face,Bad_C_I[k])) if (k!=i) break;
      if (k!=Bad_C.ne) continue;
      for (k=First_Oc[LV];k<_C->ne;k++) if (INCI_LE(New_Face,CEq_I[k]))
	if (k!=j) break;
      if (k!=_C->ne) continue;
      assert(_C->ne<CEQ_Nmax);
      CEq_I[_C->ne]=INCI_PN(INCI_D2(New_Face),0);
      _C->e[_C->ne]=EEV_To_Equation(&(Bad_C.e[i]),&(_C->e[j]),
				    _P->x[_V->v[_V->nv-1]],_P->n);
      assert(IsGoodCEq(&(_C->e[_C->ne++]),_P,_V));}

#if     (SHOW_NEW_CEq)
  {time_t DATE2=time(NULL); char sm[2]={'s',0}; 
  int Rs= (int)difftime(DATE2,DATE1);if(Rs>999){Rs/=60; *sm='m';}
  printf(" done: C.ne=%d  %d%s\n",_C->ne,Rs,sm);fflush(0);}
#endif
}

int MC_Search_New_Vertex(PolyPointList *_P, CEqList *_CEq){
  int i, chosen, min_n_out=_CEq->ne, max_n_on=0, min_prod=_CEq->ne*_CEq->ne;
  for (i=0;i<_P->np;i++){
    int j, n_out=0, n_on=0, prod;
    for (j=0;j<_CEq->ne;j++){
      Long dist=Eval_Eq_on_V(&(_CEq->e[j]),_P->x[i],_P->n);
      n_out+=(dist<0);
      n_on+=(!dist);}
    prod=n_out*(_CEq->ne-n_on-n_out);
    if (n_out)  /*if ((n_on > max_n_on)||(n_on==max_n_on)&&(n_out<min_n_out))*/
      if (prod<min_prod)
      {min_n_out=n_out; max_n_on=n_on; min_prod=prod; chosen=i;}  }
return chosen;
}

int Simple_Search_New_Vertex(PolyPointList *_P, VertexNumList *_V){
  int i;
  /*for (i=0;i<_P->np;i++){*/ 
  for (i=_P->np-1;i>=0;i--){
    int j;
    for (j=0;j<_V->nv;j++) if (i==_V->v[j]) break;
    if (j==_V->nv) return i;}
}

int  BP_Finish_Find_Equations(PolyPointList *_P, VertexNumList *_V, 
		     EqList *_F, CEqList *_CEq, INCI *F_I, INCI *CEq_I){
  int i,IP=1, First_Oc[VERT_Nmax];
  for (i=0;i<_V->nv;i++) First_Oc[i]=0;
  while(_V->nv<_P->np){ 
    assert(_V->nv<VERT_Nmax);
    /* _V->v[_V->nv++]=MC_Search_New_Vertex(_P,_CEq); */ 
    _V->v[_V->nv++]=Simple_Search_New_Vertex(_P,_V);
#if FIRST_OC
    FO_GBU_Make_New_CEqs(_P,_V,_CEq,CEq_I,First_Oc); }
#else
    GBU_Make_New_CEqs(_P,_V,_CEq,CEq_I); }
#endif
  _F->ne=_CEq->ne;
  for (i=0;i<_F->ne;i++) _F->e[i]=_CEq->e[i];
  return IP;					   
}

int  BP_Find_Equations(PolyPointList *_P, VertexNumList *_V, EqList *_F){
  /* return: IP, finds Vertices and Equations for _P even if not IP */

  STATIC CEqList CEq; STATIC INCI CEq_I[CEQ_Nmax];
  int i; INCI F_I[EQUA_Nmax];  
  if (GLZ_Start_Simplex(_P, _V, (CEqList *) _F)) return 0;
  CEq.ne=_F->ne; for(i=0;i<_F->ne;i++) CEq.e[i]=_F->e[i]; _F->ne=0;

  for (i=0;i<CEq.ne;i++) 
    if(INCI_abs(CEq_I[i]=Eq_To_INCI(&(CEq.e[i]),_P,_V))<_P->n)
      {fprintf(outFILE,"Bad CEq in Find_Equations"); exit(0);}
  return BP_Finish_Find_Equations(_P, _V, _F, &CEq, F_I, CEq_I);
}

void  PrintUsage(char *c){
  int i;
  char opt_string[28][37]={
  "f  use as filter                    ",
  "g  general output:                  ",
  "   P reflexive: Numbers of (dual)   ",
  "     points/vertices, Hodge numbers ",
  "   P not reflexive: Numbers of      ",
  "     points, vertices, equations    ",
  "p  points of P                      ",               
  "v  vertices of P                    ",
  "e  equations of P/vertices of P-dual",
  "m  pairing matrix between vertices  ", 
  "     and equations                  ",
  "d  points of P-dual                 ", 
  "     (only if P reflexive)          ",
  "a  all of the above except f        ",
  "s  check for Span property          ",
  "     (only if P from CWS)           ",
  "i  check for IP property            ",
  "I  print FaceInfo                   ",
  "w  LG-`Hodge numbers' from single   ",
  "     weight input                   ",
  "r  ignore non-reflexive input       ",
  "n  do not complete polytope or      ", 
  "     calculate Hodge numbers        ",
  "t  test stuff related to normal form",
  "S  number of symmetries             ",
  "N  normal form                      ",
  "h  print this information           ",
  "                                    "};

  printf("This is ``%s'':  computing data of a polytope P\n",c);
  puts("Usage:");
  printf("%s [-[h][f][g][w][s][i][I][p][v][e][m][d][a][r][n][t][S][N]]\n", c);
  puts("                                                [in-file [out-file]]");
  printf("Options:\n");
  for (i=0;i<14;i++) printf("%s  |  %s\n",opt_string[i],opt_string[i+14]);
  puts("");
  puts("Input:    degrees and weights `d1 w11 w12 ... d2 w21 w22 ...'");
  puts("          or `d np' or `np d' (d=Dimension, np=#[points]) and");
  puts("              (after newline) np*d coordinates"); 
  puts("Output:   as specified by options");

/*puts("Options:  h           print this information");
  puts("          f           use as filter"); 
  puts("          w           LG-`Hodge numbers' from single weight input");
  puts("          s           check for Span property (only if P from CWS)");
  puts("          i           check for IP property");
  puts("          I           Print FaceInfo");
  puts("          p           points of P"); 
  puts("          v           vertices of P"); 
  puts("          e           equations of P/vertices of P-dual"); 
  puts("          m           pairing matrix between vertices and equations"); 
  puts("          d           points of P-dual (only if P reflexive)"); 
  puts("          a           all of the above except h, f and w");
  puts("          r           ignore non-reflexive input");
  puts("          n           do not complete polytope or calculated Hodges");
  puts("          t           test stuff related to normal form");
  puts("          S           number of symmetries");
  puts("          N           normal form");

  puts("");
  puts("Input:    degrees and weights `d1 w11 w12 ... d2 w21 w22 ...'");
  puts("          or `d np' or `np d' (d=Dimension, np=#[points]) and");
  puts("              (after newline) np*d coordinates"); 
  puts("Output:   P reflexive:");
  puts("               Numbers of (dual) points/vertices, Hodge numbers");
  puts("          P not reflexive:");
  puts("               Numbers of points, vertices, equations");
  puts("          Followed by extra output specified by options");*/
}

main (int narg, char* fn[]){
  int n=0, k, FilterFlag=0, w=0, s=0, i=0, m=0, p=0, v=0, e=0, d=0, t=0,
    S=0, N=0, I=0, r=0, nc=0, g=0, IP, R, T;
  char c;
STATIC  CWS CW;
  VertexNumList V, DV; 
STATIC    EqList E, DE; 
  BaHo BH; 
  PolyPointList *_P= (PolyPointList *) malloc(sizeof (PolyPointList));
  PolyPointList *_PD = (PolyPointList *) malloc(sizeof (PolyPointList));
  if(_P==NULL) {puts("Unable to allocate space for _P"); exit(0);}
  if(_PD==NULL) {printf("Unable to allocate _PD\n"); exit(0);}
  CW.nw=0;

  while(narg > ++n) {
    if(fn[n][0]!='-') break; 
    k=0;
    while ((c=fn[n][++k])!='\0'){
      if(c=='h') { PrintUsage(fn[0]); exit(0);}
      if(c=='f') FilterFlag=1; 
      if(c=='g') g=1;
      if(c=='w') w=1;
      if(c=='s') s=1;
      if(c=='i') i=1;   
      if(c=='I') I=1;   
      if(c=='m') m=1;   
      if(c=='p') p=1;
      if(c=='v') v=1;
      if(c=='e') e=1;
      if(c=='d') d=1;
      if(c=='t') t=1;
      if(c=='S') S=1;
      if(c=='N') N=1;
      if(c=='r') r=1;
      if(c=='n') nc=1;
      if(c=='a') {g=1; m=1; p=1; v=1; e=1; d=1; }}}
  n--;
     
  if(FilterFlag) {inFILE=NULL; outFILE=stdout;}
  else {  
    if (narg > ++n)  inFILE=fopen(fn[n],"r");  
    else inFILE=stdin;
    if (inFILE==NULL){printf("Input file %s not found!\n",fn[n]);exit(0);}
    if (narg > ++n) outFILE=fopen(fn[n],"w");  
    else outFILE=stdout;     }
  
  while(Read_CWS_PP(&CW,_P)) {
    _PD->np=0;
    /* if(nc) */{R=0; if (IP=BP_Find_Equations(_P,&V,&E)) R=Ref_Equations(&E);}
    /* else R=Poly_Check(_P,&V,&E,_PD,&DV,&DE,&IP,(w||CW.nw)); */
    if (r&&!R) continue;
    if(g){
      if(!R||nc){
	BH.mp=_P->np; 
	BH.mv=V.nv; 
	BH.np=0; 
	BH.nv=E.ne;}
      else RC_Calc_BaHo(_P,&V,&E,_PD,&BH);
      Print_CWH(&CW, &BH); }
    if(s&&CW.nw) if(!Span_Check(&E,&(CW.B),&_P->n)) 
      fprintf(outFILE,"No Span\n");
    if(i && !IP) fprintf(outFILE,"No IP\n");
    if(p) Print_PPL(_P,"Points of P");
    if(v) Print_VL(_P, &V, "Vertices of P");
    if(e) { char cR[]="Vertices of P-dual <-> Equations of P",
		 cN[]="Equations of P";Print_EL(&E, &_P->n, R,(R ? cR : cN));}
    if(I){STATIC FaceInfo FI; 
	Make_Incidence(_P,&V,&E,&FI); Print_FaceInfo(_P,&FI);}
    if(m) {
      STATIC Long PM[EQUA_Nmax][VERT_Nmax];
      Make_VEPM(_P,&V,&E,PM);
      Print_Matrix(PM, E.ne, V.nv,
		   "Pairing matrix of vertices and equations of P");}
    if(d&&_PD->np) Print_PPL(_PD, "Points of P-dual");
    fflush(outFILE);     }
}
